import time
import psutil
import threading
from queue import Queue

from nbox.sublime.proto.lmao_pb2 import *

import GPUtil
HAS_GPU = len(GPUtil.getGPUs())

def get_metrics_dict():
  cpu_usage = psutil.cpu_percent(percpu=True)
  # total_memory = psutil.virtual_memory().total
  data = {
    "cpu_usage": sum(cpu_usage) / len(cpu_usage),
    "memory_available (MiB)": psutil.virtual_memory().available // (1024 ** 2),
    "memory_usage (MiB)": psutil.virtual_memory().used // (1024 ** 2),
    "memory_percentage": psutil.virtual_memory().percent,
    "disk_utilisation": psutil.disk_usage('/').percent
  }
  if HAS_GPU:
    for i, gpu in enumerate(GPUtil.getGPUs()):
      data.update({
        f"gpu-{i}_usage": gpu.load,
        f"gpu-{i}_memory_available": gpu.memoryFree // (1024 ** 2),
        f"gpu-{i}_memory_usage": gpu.memoryUsed // (1024 ** 2)
      })
  return data


class SystemMetricsLogger:
  def __init__(self, dk, log_every: int = 1) -> None:
    self.dk = dk
    self.log_every = log_every

    # create a rate limiting mechanism
    self._queue = Queue()
    self._bar = threading.Barrier(2)
    def _rate_limiter(s = log_every):
      while True:
        self._bar.wait()
        time.sleep(s)
    self.rl = threading.Thread(target=_rate_limiter, daemon=True)
    self.rl.start()

    # create a thread to create metrics
    self.metrics_logger = threading.Thread(target=self._create_metrics_dict, daemon=True)
    self.metrics_logger.start()

    # So this function returns the latest cpu usage thast it has gathered from the previous call
    # and so we need to do an init fire to avoid getting 0.0
    psutil.cpu_percent(percpu=True)

  def _create_metrics_dict(self):
    while True:
      if self.dk.completed:
        break
      data = get_metrics_dict()
      self._queue.put(data)
      run_comp = self.log()
      if run_comp:
        break
      time.sleep(self.log_every)

  def log(self) -> bool:
    if self.dk.completed:
      return True
    self._bar.wait()
    items = []
    while not self._queue.empty():
      items.append(self._queue.get())
    for x in items:
      self.dk.log(x, log_type=RunLog.LogType.SYSTEM)
    return False

  def __del__(self):
    self._bar.wait()
    self._bar.wait()
    self.rl.join()
    self.metrics_logger.join()


if __name__ == "__main__":
  import json
  print(json.dumps(get_metrics_dict(), indent=2))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: serve.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import job_pb2 as job__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bserve.proto\x12\x06\x64\x65ploy\x1a\x1bgoogle/protobuf/empty.proto\x1a google/protobuf/field_mask.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\tjob.proto\"\xed\x01\n\x07Serving\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1d\n\x06models\x18\x03 \x03(\x0b\x32\r.deploy.Model\x12-\n\x13\x64\x65ployment_endpoint\x18\x04 \x01(\x0b\x32\x10.deploy.Endpoint\x12)\n\x0fmodel_endpoints\x18\x05 \x03(\x0b\x32\x10.deploy.Endpoint\x12 \n\x08resource\x18\x06 \x01(\x0b\x32\x0e.jobs.Resource\x12-\n\x10\x63pu_scale_config\x18\x07 \x01(\x0b\x32\x13.deploy.ScaleConfig\"7\n\x08\x45ndpoint\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x12\n\naccess_key\x18\x03 \x01(\t\"\xd6\x01\n\rServingStatus\x12\x33\n\rupload_status\x18\x01 \x01(\x0e\x32\x1c.deploy.ServingStatus.Status\x12\x33\n\rdeploy_status\x18\x02 \x01(\x0e\x32\x1c.deploy.ServingStatus.Status\"[\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x12\n\x0eSTATUS_PENDING\x10\x01\x12\x12\n\x0eSTATUS_SUCCESS\x10\x02\x12\x11\n\rSTATUS_FAILED\x10\x03\"L\n\x0bScaleConfig\x12\x14\n\x0cmin_replicas\x18\x01 \x01(\x05\x12\x14\n\x0cmax_replicas\x18\x02 \x01(\x05\x12\x11\n\tthreshold\x18\x03 \x01(\x02\"X\n\x0eServingRequest\x12$\n\tauth_info\x18\x01 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12 \n\x07serving\x18\x02 \x01(\x0b\x32\x0f.deploy.Serving\"\x8f\x01\n\x14UpdateServingRequest\x12$\n\tauth_info\x18\x01 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12 \n\x07serving\x18\x02 \x01(\x0b\x32\x0f.deploy.Serving\x12/\n\x0bupdate_mask\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"Y\n\x12ServingListRequest\x12$\n\tauth_info\x18\x01 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12\r\n\x05limit\x18\x02 \x01(\x05\x12\x0e\n\x06offset\x18\x03 \x01(\x05\"K\n\x13ServingListResponse\x12!\n\x08servings\x18\x01 \x03(\x0b\x32\x0f.deploy.Serving\x12\x11\n\tnext_page\x18\x02 \x01(\x05\"\xec\x03\n\x05Model\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x18\n\x10serving_group_id\x18\x03 \x01(\t\x12.\n\ncreated_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x04type\x18\x05 \x01(\x0e\x32\x19.deploy.Model.ServingType\x12 \n\x08resource\x18\x06 \x01(\x0b\x32\x0e.jobs.Resource\x12\x1c\n\x04\x63ode\x18\x07 \x01(\x0b\x32\x0e.jobs.Job.Code\x12\"\n\x08\x65ndpoint\x18\x08 \x01(\x0b\x32\x10.deploy.Endpoint\x12+\n\npin_status\x18\t \x01(\x0e\x32\x17.deploy.Model.PinStatus\x12%\n\x06status\x18\x0e \x01(\x0b\x32\x15.deploy.ServingStatus\"E\n\x0bServingType\x12\x1c\n\x18SERVING_TYPE_UNSPECIFIED\x10\x00\x12\x18\n\x14SERVING_TYPE_NBOX_OP\x10\x01\"W\n\tPinStatus\x12\x1a\n\x16PIN_STATUS_UNSPECIFIED\x10\x00\x12\x15\n\x11PIN_STATUS_PINNED\x10\x01\x12\x17\n\x13PIN_STATUS_UNPINNED\x10\x02\"R\n\x0cModelRequest\x12$\n\tauth_info\x18\x01 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12\x1c\n\x05model\x18\x02 \x01(\x0b\x32\r.deploy.Model\"k\n\x10ModelListRequest\x12$\n\tauth_info\x18\x01 \x01(\x0b\x32\x11.jobs.NBXAuthInfo\x12\x12\n\nserving_id\x18\x02 \x01(\t\x12\r\n\x05limit\x18\x03 \x01(\x05\x12\x0e\n\x06offset\x18\x04 \x01(\x05\"E\n\x11ModelListResponse\x12\x1d\n\x06models\x18\x01 \x03(\x0b\x32\r.deploy.Model\x12\x11\n\tnext_page\x18\x02 \x01(\x05\"$\n\x08ModelLog\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04logs\x18\x02 \x03(\t2\xc4\x02\n\x0eServingService\x12\x31\n\x06\x43reate\x12\x16.deploy.ServingRequest\x1a\x0f.deploy.Serving\x12\x35\n\nGetServing\x12\x16.deploy.ServingRequest\x1a\x0f.deploy.Serving\x12G\n\x0cListServings\x12\x1a.deploy.ServingListRequest\x1a\x1b.deploy.ServingListResponse\x12?\n\rDeleteServing\x12\x16.deploy.ServingRequest\x1a\x16.google.protobuf.Empty\x12>\n\rUpdateServing\x12\x1c.deploy.UpdateServingRequest\x1a\x0f.deploy.Serving2\x9d\x03\n\x0cModelService\x12\x36\n\x05Watch\x12\x14.deploy.ModelRequest\x1a\x15.deploy.ServingStatus0\x01\x12\x32\n\x0bUploadModel\x12\x14.deploy.ModelRequest\x1a\r.deploy.Model\x12-\n\x06\x44\x65ploy\x12\x14.deploy.ModelRequest\x1a\r.deploy.Model\x12\x35\n\tModelLogs\x12\x14.deploy.ModelRequest\x1a\x10.deploy.ModelLog0\x01\x12\x41\n\nListModels\x12\x18.deploy.ModelListRequest\x1a\x19.deploy.ModelListResponse\x12;\n\x0b\x44\x65leteModel\x12\x14.deploy.ModelRequest\x1a\x16.google.protobuf.Empty\x12;\n\x0bSetModelPin\x12\x14.deploy.ModelRequest\x1a\x16.google.protobuf.EmptyB4Z2github.com/NimbleBoxAI/jobs-architecture/deploy_pbb\x06proto3')



_SERVING = DESCRIPTOR.message_types_by_name['Serving']
_ENDPOINT = DESCRIPTOR.message_types_by_name['Endpoint']
_SERVINGSTATUS = DESCRIPTOR.message_types_by_name['ServingStatus']
_SCALECONFIG = DESCRIPTOR.message_types_by_name['ScaleConfig']
_SERVINGREQUEST = DESCRIPTOR.message_types_by_name['ServingRequest']
_UPDATESERVINGREQUEST = DESCRIPTOR.message_types_by_name['UpdateServingRequest']
_SERVINGLISTREQUEST = DESCRIPTOR.message_types_by_name['ServingListRequest']
_SERVINGLISTRESPONSE = DESCRIPTOR.message_types_by_name['ServingListResponse']
_MODEL = DESCRIPTOR.message_types_by_name['Model']
_MODELREQUEST = DESCRIPTOR.message_types_by_name['ModelRequest']
_MODELLISTREQUEST = DESCRIPTOR.message_types_by_name['ModelListRequest']
_MODELLISTRESPONSE = DESCRIPTOR.message_types_by_name['ModelListResponse']
_MODELLOG = DESCRIPTOR.message_types_by_name['ModelLog']
_SERVINGSTATUS_STATUS = _SERVINGSTATUS.enum_types_by_name['Status']
_MODEL_SERVINGTYPE = _MODEL.enum_types_by_name['ServingType']
_MODEL_PINSTATUS = _MODEL.enum_types_by_name['PinStatus']
Serving = _reflection.GeneratedProtocolMessageType('Serving', (_message.Message,), {
  'DESCRIPTOR' : _SERVING,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.Serving)
  })
_sym_db.RegisterMessage(Serving)

Endpoint = _reflection.GeneratedProtocolMessageType('Endpoint', (_message.Message,), {
  'DESCRIPTOR' : _ENDPOINT,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.Endpoint)
  })
_sym_db.RegisterMessage(Endpoint)

ServingStatus = _reflection.GeneratedProtocolMessageType('ServingStatus', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGSTATUS,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ServingStatus)
  })
_sym_db.RegisterMessage(ServingStatus)

ScaleConfig = _reflection.GeneratedProtocolMessageType('ScaleConfig', (_message.Message,), {
  'DESCRIPTOR' : _SCALECONFIG,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ScaleConfig)
  })
_sym_db.RegisterMessage(ScaleConfig)

ServingRequest = _reflection.GeneratedProtocolMessageType('ServingRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGREQUEST,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ServingRequest)
  })
_sym_db.RegisterMessage(ServingRequest)

UpdateServingRequest = _reflection.GeneratedProtocolMessageType('UpdateServingRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESERVINGREQUEST,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.UpdateServingRequest)
  })
_sym_db.RegisterMessage(UpdateServingRequest)

ServingListRequest = _reflection.GeneratedProtocolMessageType('ServingListRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGLISTREQUEST,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ServingListRequest)
  })
_sym_db.RegisterMessage(ServingListRequest)

ServingListResponse = _reflection.GeneratedProtocolMessageType('ServingListResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGLISTRESPONSE,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ServingListResponse)
  })
_sym_db.RegisterMessage(ServingListResponse)

Model = _reflection.GeneratedProtocolMessageType('Model', (_message.Message,), {
  'DESCRIPTOR' : _MODEL,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.Model)
  })
_sym_db.RegisterMessage(Model)

ModelRequest = _reflection.GeneratedProtocolMessageType('ModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _MODELREQUEST,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ModelRequest)
  })
_sym_db.RegisterMessage(ModelRequest)

ModelListRequest = _reflection.GeneratedProtocolMessageType('ModelListRequest', (_message.Message,), {
  'DESCRIPTOR' : _MODELLISTREQUEST,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ModelListRequest)
  })
_sym_db.RegisterMessage(ModelListRequest)

ModelListResponse = _reflection.GeneratedProtocolMessageType('ModelListResponse', (_message.Message,), {
  'DESCRIPTOR' : _MODELLISTRESPONSE,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ModelListResponse)
  })
_sym_db.RegisterMessage(ModelListResponse)

ModelLog = _reflection.GeneratedProtocolMessageType('ModelLog', (_message.Message,), {
  'DESCRIPTOR' : _MODELLOG,
  '__module__' : 'serve_pb2'
  # @@protoc_insertion_point(class_scope:deploy.ModelLog)
  })
_sym_db.RegisterMessage(ModelLog)

_SERVINGSERVICE = DESCRIPTOR.services_by_name['ServingService']
_MODELSERVICE = DESCRIPTOR.services_by_name['ModelService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2github.com/NimbleBoxAI/jobs-architecture/deploy_pb'
  _SERVING._serialized_start=131
  _SERVING._serialized_end=368
  _ENDPOINT._serialized_start=370
  _ENDPOINT._serialized_end=425
  _SERVINGSTATUS._serialized_start=428
  _SERVINGSTATUS._serialized_end=642
  _SERVINGSTATUS_STATUS._serialized_start=551
  _SERVINGSTATUS_STATUS._serialized_end=642
  _SCALECONFIG._serialized_start=644
  _SCALECONFIG._serialized_end=720
  _SERVINGREQUEST._serialized_start=722
  _SERVINGREQUEST._serialized_end=810
  _UPDATESERVINGREQUEST._serialized_start=813
  _UPDATESERVINGREQUEST._serialized_end=956
  _SERVINGLISTREQUEST._serialized_start=958
  _SERVINGLISTREQUEST._serialized_end=1047
  _SERVINGLISTRESPONSE._serialized_start=1049
  _SERVINGLISTRESPONSE._serialized_end=1124
  _MODEL._serialized_start=1127
  _MODEL._serialized_end=1619
  _MODEL_SERVINGTYPE._serialized_start=1461
  _MODEL_SERVINGTYPE._serialized_end=1530
  _MODEL_PINSTATUS._serialized_start=1532
  _MODEL_PINSTATUS._serialized_end=1619
  _MODELREQUEST._serialized_start=1621
  _MODELREQUEST._serialized_end=1703
  _MODELLISTREQUEST._serialized_start=1705
  _MODELLISTREQUEST._serialized_end=1812
  _MODELLISTRESPONSE._serialized_start=1814
  _MODELLISTRESPONSE._serialized_end=1883
  _MODELLOG._serialized_start=1885
  _MODELLOG._serialized_end=1921
  _SERVINGSERVICE._serialized_start=1924
  _SERVINGSERVICE._serialized_end=2248
  _MODELSERVICE._serialized_start=2251
  _MODELSERVICE._serialized_end=2664
# @@protoc_insertion_point(module_scope)



"""
Here are some operators that can be used to create different architectural formats.
"""

from typing import Dict, List
from nbox import Operator

class StepOp(Operator):
  def __init__(self):
    """Convinience operator, add a no output operator using ``.add_step`` and don't write forward

    Usage
    -----

    .. code-block:: python

    class InstallPython(StepOp)
      def __init__(self, version: str = "3.9):
        self.add_step(ShellCommand(f"chmod +x ./scripts/python{version}_install.sh"))
        self.add_step(ShellCommand(f"./scripts/python{version}_install.sh"))

    install_python = InstallPython() # init op
    install_python() # call it without defining the forward function
    """
    super().__init__()
    self.steps = []

  def add_step(self, step: Operator):
    self.steps.append(step)

  def forward(self):
    for step in self.steps:
      step()

class Sequential():
  def __init__(self, *ops):
    """Package a list of operators into a sequential pipeline."""
    super().__init__()
    for op in ops:
      assert isinstance(op, Operator), "Operator must be of type Operator"
    self.ops = ops

  def forward(self, x = None, capture_output = False):
    out = x
    outputs = []
    for op in self.ops:
      out = op(out)
      outputs.append(out)
    if capture_output:
      return out, outputs
    return out



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.field_mask_pb2
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import job_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Serving(google.protobuf.message.Message):
    """
    A Serving (Group) is a collection of Model(s).
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    MODELS_FIELD_NUMBER: builtins.int
    DEPLOYMENT_ENDPOINT_FIELD_NUMBER: builtins.int
    MODEL_ENDPOINTS_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    CPU_SCALE_CONFIG_FIELD_NUMBER: builtins.int
    id: typing.Text
    """ID for the Serving Group"""

    name: typing.Text
    """name of the Serving Group"""

    @property
    def models(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Model]:
        """list of models"""
        pass
    @property
    def deployment_endpoint(self) -> global___Endpoint:
        """Endpoint for the Serving Group"""
        pass
    @property
    def model_endpoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Endpoint]:
        """Endpoints for the models. Copy of models[*].endpoint."""
        pass
    @property
    def resource(self) -> job_pb2.Resource: ...
    @property
    def cpu_scale_config(self) -> global___ScaleConfig: ...
    def __init__(self,
        *,
        id: typing.Text = ...,
        name: typing.Text = ...,
        models: typing.Optional[typing.Iterable[global___Model]] = ...,
        deployment_endpoint: typing.Optional[global___Endpoint] = ...,
        model_endpoints: typing.Optional[typing.Iterable[global___Endpoint]] = ...,
        resource: typing.Optional[job_pb2.Resource] = ...,
        cpu_scale_config: typing.Optional[global___ScaleConfig] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["cpu_scale_config",b"cpu_scale_config","deployment_endpoint",b"deployment_endpoint","resource",b"resource"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["cpu_scale_config",b"cpu_scale_config","deployment_endpoint",b"deployment_endpoint","id",b"id","model_endpoints",b"model_endpoints","models",b"models","name",b"name","resource",b"resource"]) -> None: ...
global___Serving = Serving

class Endpoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    URL_FIELD_NUMBER: builtins.int
    ACCESS_KEY_FIELD_NUMBER: builtins.int
    id: typing.Text
    """Model or Serving (Group) ID"""

    url: typing.Text
    """URL to access the model or serving group"""

    access_key: typing.Text
    """access_key is used to authenticate requests to the endpoint.
    It is used in the following ways:
    curl -H "NBX-KEY:  $access_key" https://api.nimblebox.ai/ |  $url
    """

    def __init__(self,
        *,
        id: typing.Text = ...,
        url: typing.Text = ...,
        access_key: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["access_key",b"access_key","id",b"id","url",b"url"]) -> None: ...
global___Endpoint = Endpoint

class ServingStatus(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _Status:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _StatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ServingStatus._Status.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        STATUS_UNSPECIFIED: ServingStatus._Status.ValueType  # 0
        """Unknown status"""

        STATUS_PENDING: ServingStatus._Status.ValueType  # 1
        """Deployment is pending"""

        STATUS_SUCCESS: ServingStatus._Status.ValueType  # 2
        """Deployment is successful"""

        STATUS_FAILED: ServingStatus._Status.ValueType  # 3
        """Deployment is failed"""

    class Status(_Status, metaclass=_StatusEnumTypeWrapper):
        pass

    STATUS_UNSPECIFIED: ServingStatus.Status.ValueType  # 0
    """Unknown status"""

    STATUS_PENDING: ServingStatus.Status.ValueType  # 1
    """Deployment is pending"""

    STATUS_SUCCESS: ServingStatus.Status.ValueType  # 2
    """Deployment is successful"""

    STATUS_FAILED: ServingStatus.Status.ValueType  # 3
    """Deployment is failed"""


    UPLOAD_STATUS_FIELD_NUMBER: builtins.int
    DEPLOY_STATUS_FIELD_NUMBER: builtins.int
    upload_status: global___ServingStatus.Status.ValueType
    deploy_status: global___ServingStatus.Status.ValueType
    def __init__(self,
        *,
        upload_status: global___ServingStatus.Status.ValueType = ...,
        deploy_status: global___ServingStatus.Status.ValueType = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["deploy_status",b"deploy_status","upload_status",b"upload_status"]) -> None: ...
global___ServingStatus = ServingStatus

class ScaleConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    MIN_REPLICAS_FIELD_NUMBER: builtins.int
    MAX_REPLICAS_FIELD_NUMBER: builtins.int
    THRESHOLD_FIELD_NUMBER: builtins.int
    min_replicas: builtins.int
    max_replicas: builtins.int
    threshold: builtins.float
    def __init__(self,
        *,
        min_replicas: builtins.int = ...,
        max_replicas: builtins.int = ...,
        threshold: builtins.float = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["max_replicas",b"max_replicas","min_replicas",b"min_replicas","threshold",b"threshold"]) -> None: ...
global___ScaleConfig = ScaleConfig

class ServingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTH_INFO_FIELD_NUMBER: builtins.int
    SERVING_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> job_pb2.NBXAuthInfo: ...
    @property
    def serving(self) -> global___Serving: ...
    def __init__(self,
        *,
        auth_info: typing.Optional[job_pb2.NBXAuthInfo] = ...,
        serving: typing.Optional[global___Serving] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","serving",b"serving"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","serving",b"serving"]) -> None: ...
global___ServingRequest = ServingRequest

class UpdateServingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTH_INFO_FIELD_NUMBER: builtins.int
    SERVING_FIELD_NUMBER: builtins.int
    UPDATE_MASK_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> job_pb2.NBXAuthInfo: ...
    @property
    def serving(self) -> global___Serving: ...
    @property
    def update_mask(self) -> google.protobuf.field_mask_pb2.FieldMask: ...
    def __init__(self,
        *,
        auth_info: typing.Optional[job_pb2.NBXAuthInfo] = ...,
        serving: typing.Optional[global___Serving] = ...,
        update_mask: typing.Optional[google.protobuf.field_mask_pb2.FieldMask] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","serving",b"serving","update_mask",b"update_mask"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","serving",b"serving","update_mask",b"update_mask"]) -> None: ...
global___UpdateServingRequest = UpdateServingRequest

class ServingListRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTH_INFO_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> job_pb2.NBXAuthInfo: ...
    limit: builtins.int
    offset: builtins.int
    def __init__(self,
        *,
        auth_info: typing.Optional[job_pb2.NBXAuthInfo] = ...,
        limit: builtins.int = ...,
        offset: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","limit",b"limit","offset",b"offset"]) -> None: ...
global___ServingListRequest = ServingListRequest

class ServingListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SERVINGS_FIELD_NUMBER: builtins.int
    NEXT_PAGE_FIELD_NUMBER: builtins.int
    @property
    def servings(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Serving]: ...
    next_page: builtins.int
    def __init__(self,
        *,
        servings: typing.Optional[typing.Iterable[global___Serving]] = ...,
        next_page: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["next_page",b"next_page","servings",b"servings"]) -> None: ...
global___ServingListResponse = ServingListResponse

class Model(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _ServingType:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _ServingTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Model._ServingType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        SERVING_TYPE_UNSPECIFIED: Model._ServingType.ValueType  # 0
        """Unknown deployment type"""

        SERVING_TYPE_NBOX_OP: Model._ServingType.ValueType  # 1
        """Serving via nbox Operators"""

    class ServingType(_ServingType, metaclass=_ServingTypeEnumTypeWrapper):
        """Deployment type"""
        pass

    SERVING_TYPE_UNSPECIFIED: Model.ServingType.ValueType  # 0
    """Unknown deployment type"""

    SERVING_TYPE_NBOX_OP: Model.ServingType.ValueType  # 1
    """Serving via nbox Operators"""


    class _PinStatus:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _PinStatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Model._PinStatus.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        PIN_STATUS_UNSPECIFIED: Model._PinStatus.ValueType  # 0
        """Unknown pin status"""

        PIN_STATUS_PINNED: Model._PinStatus.ValueType  # 1
        """Model is pinned"""

        PIN_STATUS_UNPINNED: Model._PinStatus.ValueType  # 2
        """Model is unpinned"""

    class PinStatus(_PinStatus, metaclass=_PinStatusEnumTypeWrapper):
        pass

    PIN_STATUS_UNSPECIFIED: Model.PinStatus.ValueType  # 0
    """Unknown pin status"""

    PIN_STATUS_PINNED: Model.PinStatus.ValueType  # 1
    """Model is pinned"""

    PIN_STATUS_UNPINNED: Model.PinStatus.ValueType  # 2
    """Model is unpinned"""


    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    SERVING_GROUP_ID_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    CODE_FIELD_NUMBER: builtins.int
    ENDPOINT_FIELD_NUMBER: builtins.int
    PIN_STATUS_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    id: typing.Text
    """ID for the model"""

    name: typing.Text
    """name of the model"""

    serving_group_id: typing.Text
    """ID of the serving group the model belongs to"""

    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """time created"""
        pass
    type: global___Model.ServingType.ValueType
    @property
    def resource(self) -> job_pb2.Resource:
        """model resource"""
        pass
    @property
    def code(self) -> job_pb2.Job.Code: ...
    @property
    def endpoint(self) -> global___Endpoint: ...
    pin_status: global___Model.PinStatus.ValueType
    @property
    def status(self) -> global___ServingStatus: ...
    def __init__(self,
        *,
        id: typing.Text = ...,
        name: typing.Text = ...,
        serving_group_id: typing.Text = ...,
        created_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        type: global___Model.ServingType.ValueType = ...,
        resource: typing.Optional[job_pb2.Resource] = ...,
        code: typing.Optional[job_pb2.Job.Code] = ...,
        endpoint: typing.Optional[global___Endpoint] = ...,
        pin_status: global___Model.PinStatus.ValueType = ...,
        status: typing.Optional[global___ServingStatus] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["code",b"code","created_at",b"created_at","endpoint",b"endpoint","resource",b"resource","status",b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["code",b"code","created_at",b"created_at","endpoint",b"endpoint","id",b"id","name",b"name","pin_status",b"pin_status","resource",b"resource","serving_group_id",b"serving_group_id","status",b"status","type",b"type"]) -> None: ...
global___Model = Model

class ModelRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTH_INFO_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> job_pb2.NBXAuthInfo: ...
    @property
    def model(self) -> global___Model: ...
    def __init__(self,
        *,
        auth_info: typing.Optional[job_pb2.NBXAuthInfo] = ...,
        model: typing.Optional[global___Model] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","model",b"model"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","model",b"model"]) -> None: ...
global___ModelRequest = ModelRequest

class ModelListRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTH_INFO_FIELD_NUMBER: builtins.int
    SERVING_ID_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> job_pb2.NBXAuthInfo: ...
    serving_id: typing.Text
    limit: builtins.int
    offset: builtins.int
    def __init__(self,
        *,
        auth_info: typing.Optional[job_pb2.NBXAuthInfo] = ...,
        serving_id: typing.Text = ...,
        limit: builtins.int = ...,
        offset: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","limit",b"limit","offset",b"offset","serving_id",b"serving_id"]) -> None: ...
global___ModelListRequest = ModelListRequest

class ModelListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    MODELS_FIELD_NUMBER: builtins.int
    NEXT_PAGE_FIELD_NUMBER: builtins.int
    @property
    def models(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Model]: ...
    next_page: builtins.int
    def __init__(self,
        *,
        models: typing.Optional[typing.Iterable[global___Model]] = ...,
        next_page: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["models",b"models","next_page",b"next_page"]) -> None: ...
global___ModelListResponse = ModelListResponse

class ModelLog(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    LOGS_FIELD_NUMBER: builtins.int
    id: typing.Text
    """ID for the model"""

    @property
    def logs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """logs of the model"""
        pass
    def __init__(self,
        *,
        id: typing.Text = ...,
        logs: typing.Optional[typing.Iterable[typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","logs",b"logs"]) -> None: ...
global___ModelLog = ModelLog



"""
# NBX-Model

This will combine all the NimbleBox modules into a single callable.
"""


class Model:
  def __init__(
    self,
    # # model: Any,
    # # method: str = None,
    # # pre: callable = None,
    # # post: callable = None,
    # # model_spec: ModelSpec = None,
    # # verbose: bool = False
  ):
    """Top of the stack Model class."""
    raise NotImplementedError("WIP check back later.")

  def __repr__(self):
    return f"<nbox.Model: {self.model} >"

  def __call__(self, input_object):
    r"""Call is the most important UI/UX. The ``input_object`` can be anything from
    a tensor, an image file, filepath as string, string and is processed by ``pre`` function."""
    raise NotImplementedError("WIP check back later.")

  @classmethod
  def deserialise(cls):
    """Load ``ModelSpec`` and ``folder`` with the files in it and return a ``Model`` object."""
    raise NotImplementedError("WIP check back later.")


  def deploy(self,):
    """Serve your model on NBX-Deploy `read more <https://nimbleboxai.github.io/nbox/nbox.model.html>`_"""
    raise NotImplementedError("WIP check back later.")

  @staticmethod
  def train_on_instance():
    """Train this model on an NBX-Build Instance. Though this function is generic enough to execute
    any arbitrary code, this is built primarily for internal use."""
    raise NotImplementedError("WIP check nbox.Job")

  @staticmethod
  def train_on_job():
    """Train this model on NBX-Jobs. This same experience can be given by creating a new job
    that then can then be populated how we currently create a new job using NBX-Jobs CLI"""
    raise NotImplementedError("WIP check nbox.Job")




"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import google.protobuf.empty_pb2
import google.protobuf.wrappers_pb2
import grpc
import job_pb2
import nbox_ws_pb2
import typing

class WSJobServiceStub:
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService manages
    DB and NBX Platform Information. All compute resource related actions are fowarded
    to WSJobsService running on AWS.

    // Generated according to https://cloud.google.com/apis/design/standard_methods
    """
    def __init__(self, channel: grpc.Channel) -> None: ...
    GetJob: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.JobInfo,
        job_pb2.Job]
    """GetJob returns the job with the given info"""

    ListJobs: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.ListJobsRequest,
        nbox_ws_pb2.ListJobsResponse]
    """ListJobs returns a list of jobs for the given workspace/user"""

    UploadJobCode: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.UploadCodeRequest,
        job_pb2.Job]
    """To upload Job Code to S3"""

    CreateJob: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.CreateJobRequest,
        job_pb2.Job]
    """CreateJob creates a new job"""

    UpdateJob: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.UpdateJobRequest,
        job_pb2.Job]
    """UpdateJob updates the given job"""

    DeleteJob: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.JobInfo,
        google.protobuf.empty_pb2.Empty]
    """DeleteJob deletes the given job"""

    GetJobLogs: grpc.UnaryStreamMultiCallable[
        nbox_ws_pb2.JobLogsRequest,
        nbox_ws_pb2.JobLog]
    """GetJobLogs returns logs from given job"""

    TriggerJob: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.JobInfo,
        google.protobuf.wrappers_pb2.BoolValue]
    """TriggerJob triggers the given job"""

    UpdateRun: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.UpdateRunRequest,
        google.protobuf.empty_pb2.Empty]
    """UpdateRun update the run for given job"""

    GetRunToken: grpc.UnaryUnaryMultiCallable[
        nbox_ws_pb2.JobInfo,
        google.protobuf.wrappers_pb2.StringValue]
    """GetRunToken returns a tokn to be used for `/UpdateRun`"""


class WSJobServiceServicer(metaclass=abc.ABCMeta):
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService manages
    DB and NBX Platform Information. All compute resource related actions are fowarded
    to WSJobsService running on AWS.

    // Generated according to https://cloud.google.com/apis/design/standard_methods
    """
    @abc.abstractmethod
    def GetJob(self,
        request: nbox_ws_pb2.JobInfo,
        context: grpc.ServicerContext,
    ) -> job_pb2.Job:
        """GetJob returns the job with the given info"""
        pass

    @abc.abstractmethod
    def ListJobs(self,
        request: nbox_ws_pb2.ListJobsRequest,
        context: grpc.ServicerContext,
    ) -> nbox_ws_pb2.ListJobsResponse:
        """ListJobs returns a list of jobs for the given workspace/user"""
        pass

    @abc.abstractmethod
    def UploadJobCode(self,
        request: nbox_ws_pb2.UploadCodeRequest,
        context: grpc.ServicerContext,
    ) -> job_pb2.Job:
        """To upload Job Code to S3"""
        pass

    @abc.abstractmethod
    def CreateJob(self,
        request: nbox_ws_pb2.CreateJobRequest,
        context: grpc.ServicerContext,
    ) -> job_pb2.Job:
        """CreateJob creates a new job"""
        pass

    @abc.abstractmethod
    def UpdateJob(self,
        request: nbox_ws_pb2.UpdateJobRequest,
        context: grpc.ServicerContext,
    ) -> job_pb2.Job:
        """UpdateJob updates the given job"""
        pass

    @abc.abstractmethod
    def DeleteJob(self,
        request: nbox_ws_pb2.JobInfo,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """DeleteJob deletes the given job"""
        pass

    @abc.abstractmethod
    def GetJobLogs(self,
        request: nbox_ws_pb2.JobLogsRequest,
        context: grpc.ServicerContext,
    ) -> typing.Iterator[nbox_ws_pb2.JobLog]:
        """GetJobLogs returns logs from given job"""
        pass

    @abc.abstractmethod
    def TriggerJob(self,
        request: nbox_ws_pb2.JobInfo,
        context: grpc.ServicerContext,
    ) -> google.protobuf.wrappers_pb2.BoolValue:
        """TriggerJob triggers the given job"""
        pass

    @abc.abstractmethod
    def UpdateRun(self,
        request: nbox_ws_pb2.UpdateRunRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """UpdateRun update the run for given job"""
        pass

    @abc.abstractmethod
    def GetRunToken(self,
        request: nbox_ws_pb2.JobInfo,
        context: grpc.ServicerContext,
    ) -> google.protobuf.wrappers_pb2.StringValue:
        """GetRunToken returns a tokn to be used for `/UpdateRun`"""
        pass


def add_WSJobServiceServicer_to_server(servicer: WSJobServiceServicer, server: grpc.Server) -> None: ...



"""
``Operators`` is how you write NBX-Jobs. If you are familiar with pytorch, then usage is
exactly same, for others here's a quick recap:

.. code-block:: python

  class MyOperator(Operator):
    def __init__(self, a: int, b: str):
      super().__init__()
      self.a: int = a
      self.b: Operator = MyOtherOperator(b) # nested calling
    
    def forward(self, x: int) -> int:
      y = self.a + x
      y = self.b(y) + x # nested calling
      return y

  job = MyOperator(1, "hello") # define once
  res = job(2)                 # use like python, screw DAGs

If you want to use deploy `nbox.Jobs <nbox.jobs.html>`_ is a better documentation.


Engineering
-----------

Fundamentally operators act as a wrapper on user code, sometime abstracting away functions
by breaking them into ``__init__``s and ``forward``s. But this is a simpler way to wrap
user function than letting users wrap their own function. It is easy to get false positives,
and so we explicitly expand things in two. These operators are like ``torch.nn.Modules``
spiritually as well because modules manage the underlying weights and operators manage the
underlying user logic.

Operators are combination of several subsystems that are all added in the same class, though
certainly if we come up with that high abstraction we will refactor this:

#. tree: All operators are really treated like a tree meaning that the execution is nested\
    and the order of execution is determined by the order of the operators in the tree. DAGs\
    are fundamentally just trees with some nodes spun togeather, to execute only once.
#. deploy, ...: All the services in NBX-Jobs.
#. get_nbx_flow: which is the static code analysis system to understand true user intent and\
    if possible (and permission of the user) optimise the logic.


Tips
----

``Operators`` are built to be the abstract equivalent of any computation so code can be easily
run in distributed fashion.

#. Use Operator directly as a function as much as possible, it's the simplest way to use it.
#. ``@operator`` decorator on your function and it will be run as a job by default, you want that.
#. ``@operator`` decorator on your class and it will be run as a serving by default, you want that.

Documentation
-------------
"""
# Some parts of the code are based on the pytorch nn.Module class
# pytorch license: https://github.com/pytorch/pytorch/blob/master/LICENSE
# modifications: research@nimblebox.ai 2022

from functools import partial
import json
import os
import inspect
from time import sleep
import requests
from enum import Enum
from typing import Any, Dict, Iterable, List, Tuple, Union
from collections import OrderedDict
from subprocess import Popen

import nbox.utils as U
from nbox.utils import logger, env
from nbox.nbxlib.tracer import Tracer
from nbox.version import __version__
from nbox.sub_utils.latency import log_latency
from nbox.framework.on_functions import get_nbx_flow
from nbox.framework import AirflowMixin, PrefectMixin
from nbox.hyperloop.job_pb2 import Job as JobProto, Resource
from nbox.hyperloop.dag_pb2 import DAG, Flowchart, Node, RunStatus
from nbox.network import deploy_job, deploy_serving, _get_deployment_data, _get_job_data
from nbox.jobs import Schedule, new as new_folder, Job, Serve
from nbox.messages import get_current_timestamp, write_binary_to_file
from nbox.relics import RelicsNBX
from nbox.init import nbox_ws_v1


class OperatorType(Enum):
  """This Enum does not concern the user, however I am describing it so people can get a feel of the breadth
  of what nbox can do. The purpose of ``Operator`` is to build an abstract representation of any possible compute
  and execute them in any fashion needed to improve the overall performance of any distributed software system.
  Here are the different types of operators:

  #. ``UNSET``: this is the default mode and is like using vanilla python without any nbox features.
  #. ``JOB``: In this case the process is run as a batch process and the I/O of values is done using Relics
  #. ``SERVING``: In this case the process is run as an API proces
  #. ``WRAP_FN``: When we wrap a function as an ``Operator``, by default deployed as a job
  """
  UNSET = "unset" # default
  JOB = "job"
  SERVING = "serving"
  WRAP_FN = "fn_wrap"
  WRAP_CLS = "cls_wrap"

  def _valid_deployment_types():
    return (OperatorType.JOB.value, OperatorType.SERVING.value)


DEFAULT_RESOURCE = Resource(
  cpu = "100m",         # 100mCPU
  memory = "200Mi",     # MiB
  disk_size = "1Gi",    # GiB
  gpu = "none",         # keep "none" for no GPU
  gpu_count = "0",      # keep "0" when no GPU
  timeout = 120_000,    # 2 minutes between attempts
  max_retries = 3,      # third times the charm :P
)

def generic_method(fn, base_url, session, *args, **kwargs):
  r = session.post(f"{base_url}/forward/{fn}")


class Operator():
  node = Node()
  source_edges: List[Node] = None
  _op_type = OperatorType.UNSET
  _op_wrap = None

  # this is a map from operator to resource, by deafult the values will be None,
  # unless explicitly modified by `chief.machine.Machine`
  _current_resource = None
  _op_to_resource_map = {}

  def __init__(self) -> None:
    """Create an operator, which abstracts your code into sharable, bulding blocks which
    can then deployed on either NBX-Jobs or NBX-Deploy.
    
    Usage:

    .. code-block:: python
      
      class MyOperator(Operator):
        def __init__(self, ...):
          ... # initialisation of job happens here

          # use prebuilt operators to define the entire process
          from nbox.lib.shell import Shell
          self.download_binary = Shell("wget https://nbox.ai/{hash}")

          # keep a library of organisation wide operators
          from .nbx_internal.operators import TrainGPT
          self.train_model: Operator = TrainGPT(...)

        def forward(self, ...):
          # pass any inputs you want at runtime
          ... # execute code in any arbitrary order, free from DAGs

          self.download_binary(hash="my_binary") # pass relevant data at runtime
          self.train_model() # run any operation and get full visibility on platform

      # to convert operator is 
      job: Operator = MyOperator(...)

      # deploy this as a batch process or API endpoint or API endpoint
      job.deploy() # WIP # WIP
    """
    self._operators: Dict[str, 'Operator'] = OrderedDict() # {name: operator}
    self._op_trace = []
    self._tracer: Tracer = None

  def __remote_init__(self):
    """User can overwrite this function, this will be called only when running on remote.
    This helps in with things like creating the models can caching them in self, instead
    of ``lru_cache`` in forward."""
    pass

  def remote_init(self):
    """Triggers `__remote_init__` across the entire tree."""
    self.__remote_init__()
    for _, op in self._operators.items():
      op.remote_init()

  def __repr__(self):
    # from torch.nn.Module
    def _addindent(s_, numSpaces):
      s = s_.split('\n')
      # don't do anything for single-line stuff
      if len(s) == 1:
        return s_
      first = s.pop(0)
      s = [(numSpaces * ' ') + line for line in s]
      s = '\n'.join(s)
      s = first + '\n' + s
      return s

    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    extra_repr = ""
    # empty string will be split into list ['']
    if extra_repr:
      extra_lines = extra_repr.split('\n')
    child_lines = []
    for key, module in self._operators.items():
      mod_str = repr(module)
      mod_str = _addindent(mod_str, 2)
      child_lines.append('(' + key + '): ' + mod_str)
    lines = extra_lines + child_lines

    main_str = self.__qualname__ + '('
    if lines:
      # simple one-liner info, which most builtin Modules will use
      if len(extra_lines) == 1 and not child_lines:
        main_str += extra_lines[0]
      else:
        main_str += '\n  ' + '\n  '.join(lines) + '\n'

    main_str += ')'
    return main_str

  # mixin/

  to_airflow_operator = AirflowMixin.to_airflow_operator
  to_airflow_dag = AirflowMixin.to_airflow_dag
  from_airflow_operator = classmethod(AirflowMixin.from_airflow_operator)
  from_airflow_dag = classmethod(AirflowMixin.from_airflow_dag)
  to_prefect_task = PrefectMixin.to_prefect_task
  to_prefect_flow = PrefectMixin.to_prefect_flow
  from_prefect_task = classmethod(PrefectMixin.from_prefect_task)
  from_prefect_flow = classmethod(PrefectMixin.from_prefect_flow)

  # Operators are very versatile and can be created in many different ways, heres a list of those:
  # 1. from an existing job, where calling it triggers a new run and args are passed through relics
  # 2. from an existing serving, where calling it equates to an HTTP API call
  # 3. from a function, where we decorate an existing function and convert it to an operator

  @classmethod
  def from_job(cls, job_id_or_name, workspace_id: str):
    """latch an existing job so that it can be called as an operator."""
    # implement this when we have the client-server that allows us to get the metadata for the job
    if not workspace_id:
      raise DeprecationWarning("Personal workspace does not support serving")
    job_id, job_name = _get_job_data(job_id_or_name, workspace_id)
    if job_id is None:
      raise ValueError(f"No serving found with name {job_name}")

    logger.debug(f"Latching to job '{job_name}' ({job_id})")
    
    def forward(*args, **kwargs):
      """This is the forward method for a NBX-Job. All the parameters will be passed through Relics."""
      logger.debug(f"Running job '{job_name}' ({job_id})")
      job = Job(job_id, workspace_id)
      relic = RelicsNBX("dot_deploy_cache", workspace_id, create = True)
      
      # determining the put location is very tricky because there is no way to create a sync between the
      # key put here and what the run will pull. This can lead to many weird race conditions. So for now
      # I am going to rely on the fact that we cannot have two parallel active runs. Thus at any given
      # moment there can be only one file at /{job_id}/args_kwargs.pkl
      relic.put_object(f"{job_id}/args_kwargs", (args, kwargs))

      # and then we will trigger the job and wait for the run to complete
      job.trigger()
      latest_run = job.last_n_runs(1)
      logger.debug(f"New run ID: {latest_run['id']}")
      max_retries = latest_run['resource']['max_retries']

      max_polls = 600  # about 10 mins
      poll_count = 0
      while latest_run["retry_count"] <= max_retries:
        # create a polling loop
        logger.debug(f"[{poll_count:04d}/{max_polls:04d}] [{latest_run['retry_count']}/{max_retries}] {latest_run['status']}")
        if poll_count > max_polls:
          raise TimeoutError(f"Run {latest_run['id']} timed out after {max_polls} polls")
        if latest_run["status"] == "COMPLETED":
          break
        elif latest_run["retry_count"] == max_retries and latest_run["status"] == "ERROR":
          raise RuntimeError(f"Run {latest_run['id']} failed after {max_retries} retries")
        latest_run = job.last_n_runs(1)
        poll_count += 1
        sleep(1)

      if latest_run["status"] == "ERROR":
        raise Exception(f"Run failed after {max_retries} retries")

      # assuming everything went well we should have a file at /{job_id}/return
      obj = relic.get_object(f"{job_id}/return")
      return obj
    
    # create the class and override some values to make more sense
    _op = cls()
    _op.__qualname__ = "job_" + job_id
    _op.forward = forward
    _op._op_type = OperatorType.JOB

    return _op

  @classmethod
  def from_serving(cls, deployment_id_or_name, token: str, workspace_id: str):
    """Latch to an existing serving operator

    Args:
      deployment_id_or_name (str): The id or name of the deployment
      token (str): The token to access the deployment, get it from settings
      workspace_id (str, optional): The workspace id. Defaults to None.
    """
    if not workspace_id:
      raise DeprecationWarning("Personal workspace does not support serving")
    serving_id, serving_name = _get_deployment_data(deployment_id_or_name, workspace_id)
    if serving_id is None:
      raise ValueError(f"No serving found with name {serving_name}")
    
    logger.debug(f"Latching to serving '{serving_name}' ({serving_id})")

    # common things
    session = requests.Session()
    session.headers.update({"NBX-KEY": token})

    REQ = type("REQ", (object,), {})

    # get the OpenAPI spec for the serving
    url = f"https://api.nimblebox.ai/{serving_id}/openapi.json"
    r = session.get(url)
    data = r.json()

    # create a function to input spec mapper and so the generic forward method can be 
    fn_spec = {}
    for p, v in data["paths"].items():
      if p.startswith("/method_"):
        fn_name = p[8:]
        v = v["post"]
        op_id = v["operationId"]
        ref_path = v["requestBody"]["content"]["application/json"]["schema"]["$ref"].split("/")[1:]
        out = data
        for p in ref_path:
          out = out[p]
        args = out["properties"]
        fn_spec[fn_name] = args

    # define the forward function for this serving operator, the objective is that this will be able to handle
    # args, kwargs just like how it works on the local machine
    def forward(method, *args, **kwargs):
      print(method, args)
      if method in fn_spec:
        # this is a simple method call
        url = f"https://api.nimblebox.ai/{serving_id}/method_{method}"
        
        # check in the kwargs if we have any arguments to pass
        args_dict = {k: v.get("default", REQ) for k, v in fn_spec[method].items()}
        _data = {} # the json that will be sent over
        for i, (k, v) in enumerate(args_dict.items()):
          if len(args) == i:
            break
          _data[k] = args[i]
        for k, v in kwargs.items():
          _data[k] = v

        # client-side check for unknown vars
        unknown_args = set()
        for k, v in _data.items():
          if k not in args_dict:
            unknown_args.add(k)
        if len(unknown_args):
          raise ValueError(f"Unknown arguments: {unknown_args}")

        missing_args = set()
        for k, v in args_dict.items():
          if k not in _data and v == REQ:
            missing_args.add(k)
        if len(missing_args):
          raise ValueError(f"Missing required arguments: {missing_args}")
      else:
        # let's see if we can do something about this information from the server
        url = f"https://api.nimblebox.ai/{serving_id}/nbx_py_rpc"
        _data = {
          "rpc_name": method,
          "key": U.py_to_bs64(args[0]),
        }
        if len(args) > 1:
          _data["value"] = U.py_to_bs64(args[1])

      # make a POST call to /forward and return the json
      logger.debug(f"Hitting URL: {url}")
      logger.debug(f"Data: {_data}")
      r = session.post(url, json = _data)
      if r.status_code != 200:
        raise Exception(r.text)
      print(r.content)
      try:
        data = r.json()
      except:
        logger.error(r.text)
        return None

      if not data["success"]:
        raise Exception(data["message"])

      value = U.py_from_bs64(data["value"])
      return value
      

    # create the class and override some values to make more sense
    _op = cls()
    _op.propagate(_nbx_serving_fn_spec = fn_spec)
    _op.__qualname__ = "serving_" + serving_id
    _op.forward = forward
    _op._op_type = OperatorType.SERVING
    return _op

  @classmethod
  def fn(cls):
    """Wraps the function as an Operator, so you can use all the same methods as Operator"""
    def wrap(fn):
      if type(fn) == type(wrap): # lol the quick hack
        # this is a wrapped function to be run as a job
        op = cls()
        op = op._fn(fn)
        return op
      elif type(fn) == type(Operator):
        # this is a little bit tricky since the initialisation of the object has be be done
        # by the user later and thus we wrap another function which actually initialises the
        # object
        def cls_init(*args, **kwargs):
          op = cls()
          op = op._cls(fn, *args, **kwargs)
          return op
        return cls_init
    return wrap

  def _cls(self, fn, *args, **kwargs):
    """Do not use directly, use ``@operator`` decorator instead. Utility to wrap a class as an operator"""
    obj = fn(*args, **kwargs)
    # override the file this is necessary for getting the remote operations running
    self.__file__ = inspect.getfile(fn)
    self.__doc__ = obj.__doc__
    self.__qualname__ = "cls_" + obj.__class__.__qualname__
    self._op_type = OperatorType.WRAP_CLS
    self._op_wrap = obj
    return self

  def _fn(self, fn):
    """Do not use directly, use ``@operator`` decorator instead. Utility to wrap a function as an operator"""
    self.forward = fn # override the forward function
    # override the file this is necessary for getting the remote operations running
    self.__file__ = inspect.getfile(fn)
    self.__doc__ = fn.__doc__
    self.__qualname__ = "fn_" + fn.__name__
    self._op_type = OperatorType.WRAP_FN
    self._op_wrap = fn
    return self

  # /mixin

  # python state modification /

  """The functions below are part of a larger effort to make Operator interact better with the underlying
  user classes and functions. Most Importantly we are borrowing the ray concept that functions map to jobs
  and classes map to actors. We are also extending it a little bit to make it more pythonic in ways by adding
  suppoprt for multiple inbuilt python object methods like __getitem__, __setitem__ etc. So in order to make
  it work we are adding those methods below as well and each method will based on its type take a call on
  what to do.
  """

  def __setattr__(self, key, value: 'Operator'):
    obj = getattr(self, key, None)
    if key != "forward" and obj is not None and callable(obj) and not isinstance(value, Operator):
      raise AttributeError(f"cannot assign {key} as it is already a method")
    if isinstance(value, Operator):
      if not "_operators" in self.__dict__:
        raise AttributeError("cannot assign operator before super().__init__() call")
      if key in self.__dict__ and key not in self._operators:
        raise KeyError(f"attribute '{key}' already exists")
      self._operators[key] = value

      # map the operator to the current resource
      self._op_to_resource_map[key] = self._current_resource
    self.__dict__[key] = value

  def __getattr__(self, key):
    if self._op_type == OperatorType.SERVING:
      if key in self._nbx_serving_fn_spec:
        return partial(self.forward, key)
      return self.forward("__getattr__", key)
    elif self._op_type == OperatorType.WRAP_CLS:
      return getattr(self._op_wrap, key)
    raise AttributeError(f"{key}")

  def __getitem__(self, key):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__getitem__", key)
    if self._op_type in [OperatorType.WRAP_FN, OperatorType.WRAP_CLS]:
      return self._op_wrap[key]
    raise KeyError(f"{key}")

  def __setitem__(self, key, value):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__setitem__", key, value)
    if self._op_type in [OperatorType.WRAP_CLS]:
      self._op_wrap[key] = value
      return
    raise KeyError(f"{key}")

  def __delitem__(self, key):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__delitem__", key)
    if self._op_type in [OperatorType.WRAP_CLS]:
      del self._op_wrap[key]; return
    raise KeyError(f"{key}")

  def __iter__(self):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__iter__")
    if self._op_type in [OperatorType.WRAP_CLS]:
      return iter(self._op_wrap)
    raise ValueError(f"Operator cannot iterate")

  def __next__(self):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__next__")
    if self._op_type in [OperatorType.WRAP_CLS]:
      return next(self._op_wrap)
    raise ValueError(f"Operator cannot iterate")

  def __len__(self):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__len__")
    if self._op_type in [OperatorType.WRAP_CLS]:
      return len(self._op_wrap)
    raise ValueError(f"Operator cannot iterate")

  def __contains__(self, key):
    if self._op_type == OperatorType.SERVING:
      return self.forward("__contains__", key)
    if self._op_type in [OperatorType.WRAP_CLS]:
      return key in self._op_wrap
    raise ValueError(f"Operator cannot iterate")

  # / python state modification

  def propagate(self, **kwargs):
    """Set kwargs for each child in the Operator"""
    for k, v in kwargs.items():
      setattr(self, k, v)
    for c in self._operators.values():
      c.propagate(**kwargs)

  def thaw(self, job: JobProto):
    """Load JobProto into this Operator"""
    nodes = job.dag.flowchart.nodes
    edges = job.dag.flowchart.edges
    for _id, node in nodes.items():
      name = node.name
      if name.startswith("self."):
        name = name[5:]
      if hasattr(self, name):
        op: 'Operator' = getattr(self, name)
        op.propagate(
          node = node,
          source_edges = list(filter(
            lambda x: edges[x].target == node.id, edges.keys()
          ))
        )

  def named_operators(self, memo = None, prefix: str = '', remove_duplicate: bool = True):
    r"""Returns an iterator over all modules in the network, yielding both the name of the module
    as well as the module itself."""
    if memo is None:
      memo = set()
    if self not in memo:
      if remove_duplicate:
        memo.add(self)
      yield prefix, self
      for name, module in self._operators.items():
        if module is None:
          continue
        submodule_prefix = prefix + ('.' if prefix else '') + name
        for m in module.named_operators(memo, submodule_prefix, remove_duplicate):
          yield m

  def __call__(self, *args, **kwargs):
    # check if calling is allowed based on the type
    if self._op_type == OperatorType.SERVING:
      # the fn_spec is set during the .from_serving() classmethod
      if len(self._nbx_serving_fn_spec) == 1 and "forward" in self._nbx_serving_fn_spec:
        # this means that an OperatorType.UNSET or OperatorType.WRAP_FN are on the other side of the pipe
        return self.forward("forward", *args, **kwargs)
      else:
        # this means that OperatorType.WRAP_CLS is on the other side of the pipe
        raise ValueError(f"Cannot call servings directly, will interfere with nbox")
    elif self._op_type == OperatorType.WRAP_CLS:
      raise ValueError(f"Cannot call class wrappers directly, will interfere with nbox")

    input_dict = {}
    logger.debug(f"Calling operator '{self.__class__.__name__}': {self.node.id}")
    _ts = get_current_timestamp()
    self.node.run_status.CopyFrom(RunStatus(start = _ts, inputs = {k: str(type(v)) for k, v in input_dict.items()}))
    if self._tracer != None:
      self._tracer(self.node)
    # ---- USER SEPERATION BOUNDARY ---- #

    with log_latency(f"{self.__class__.__name__}-forward"):
      out = self.forward(*args, **kwargs)

    # ---- USER SEPERATION BOUNDARY ---- #
    outputs = {}
    logger.debug(f"Ending operator '{self.__class__.__name__}': {self.node.id}")
    _ts = get_current_timestamp()
    self.node.run_status.MergeFrom(RunStatus(end = _ts, outputs = {k: str(type(v)) for k, v in outputs.items()}))
    if self._tracer != None:
      self._tracer(self.node)
    return out

  def forward(self):
    raise NotImplementedError("User must implement forward()")

  # nbx/
  def _get_dag(self) -> DAG:
    """Get the DAG for this Operator including all the nested ones."""
    dag = get_nbx_flow(self.forward)
    all_child_nodes = {}
    all_edges = {}
    for child_id, child_node in dag.flowchart.nodes.items():
      name = child_node.name
      if name.startswith("self."):
        name = name[5:]
      operator_name = "CodeBlock" # default
      cls_item = getattr(self, name, None)
      if cls_item and cls_item.__class__.__base__ == Operator:
        # this node is an operator
        operator_name = cls_item.__class__.__name__
        child_dag: DAG = cls_item._get_dag() # call this function recursively

        # update the child nodes with parent node id
        for _child_id, _child_node in child_dag.flowchart.nodes.items():
          if _child_node.parent_node_id == "":
            _child_node.parent_node_id = child_id # if there is already a parent_node_id set is child's child
          all_child_nodes[_child_id] = _child_node
          all_edges.update(child_dag.flowchart.edges)
      
      # update the child nodes name with the operator name
      child_node.operator = operator_name

    # update the root dag with new children and edges
    _nodes = {k:v for k,v in dag.flowchart.nodes.items()}
    _edges = {k:v for k,v in dag.flowchart.edges.items()}
    _nodes.update(all_child_nodes)
    _edges.update(all_edges)

    # because updating map in protobuf is hard
    dag.flowchart.CopyFrom(Flowchart(nodes = _nodes, edges = _edges))
    return dag

  def deploy(
    self,
    workspace_id: str,
    id_or_name:str = None,
    deployment_type: str = None,
    resource: Resource = None,
    *,
    _unittest = False
  ) -> 'Operator':
    """Uploads relevant files to the cloud and deploys as a batch process or and API endpoint, returns the relevant
    ``.from_job()`` or ``.from_serving`` Operator. This uploads the entire folder where the caller file is located.
    In which case having a ``.nboxignore`` and ``requirements.txt`` will also be moved over.

    Args:
      workspace_id (str): The workspace id to deploy to.
      id_or_name (str, optional): The id or name of the deployment. if deployment_type is 'serving' this this must be provided.
      deployment_type (str, optional): The type of deployment, should be one of 'serving', 'job'. Defaults to 'job'.
      resource (Resource, optional): The resource to deploy to, uses a reasonable default.
    """
    if deployment_type not in OperatorType._valid_deployment_types():
      raise ValueError(f"Invalid deployment type: {deployment_type}. Must be one of {OperatorType._valid_deployment_types()}")
    if deployment_type == OperatorType.SERVING.value:
      if id_or_name is None:
        raise ValueError("id_or_name must be provided for serving deployment")

    if self._op_type == OperatorType.UNSET:
      fp = inspect.getfile(self.__class__)
      name = self.__qualname__
    elif self._op_type in [OperatorType.JOB, OperatorType.SERVING]:
      raise ValueError("Cannot deploy an operator that is already deployed")
    elif self._op_type == OperatorType.WRAP_FN:
      fp = self.__file__
      name = self.__qualname__[3:] # to account for "fn_"
    elif self._op_type == OperatorType.WRAP_CLS:
      fp = self.__file__
      name = self.__qualname__[4:] # to account for "cls_"

    logger.info(f"Deploying '{name}' from '{fp}'")
    logger.info(f"Deployment Type: {deployment_type}")

    # so basically till now all we know is that `from fp import name` and `name()`. This process can now
    # be used to create a custom `nbx_user.py` file from which `exe.py` can import the three functions
    # and use them the conventional way.

    # create a temporary directory to store all the files
    nbx_folder = os.path.join(env.NBOX_HOME_DIR(), ".auto_dep", name)
    new_folder(nbx_folder)

    # just create a resource.pb, if it's empty protobuf will work it out
    rsp = os.path.join(nbx_folder, "resource.pb")
    write_binary_to_file(resource or DEFAULT_RESOURCE, file = rsp)

    # copy over all the files and wait for it, else the changes below won't be reflected
    folder, file = os.path.split(fp)
    Popen(f'cp -r {folder}/ {nbx_folder}/', shell = True).wait()

    # in a way nbx_user.py and the three functions could simply be a router and the real functions can be
    # imported from the orginal file.
    with open(os.path.join(nbx_folder, "nbx_user.py"), "w") as f:
      f.write(f'''# Autogenerated for .deploy() call
from nbox.messages import read_file_to_binary
from nbox.hyperloop.job_pb2 import Resource

from {file.split('.')[0]} import {name}

def get_op(*_, **__):
  # takes no input since programtically generated returns the exact object
  out = {name}{
    '()' if self._op_type in [OperatorType.UNSET, OperatorType.WRAP_CLS] else ''
  }
  return out

get_resource = lambda: read_file_to_binary('{rsp}', message = Resource())
get_schedule = lambda: None
''')

    if _unittest:
      return

    if deployment_type == OperatorType.JOB.value:
      out = Job.upload(
        init_folder = nbx_folder,
        id_or_name = id_or_name or "dot_deploy_runs", # all the .deploy() methods will be called dot_deploy_runs
        workspace_id = workspace_id,
      )
      return self.from_job(
        job_id_or_name = out.id,
        workspace_id = workspace_id,
      )
    elif deployment_type == OperatorType.SERVING.value:
      out = Serve.upload(
        init_folder = nbx_folder,
        id_or_name = id_or_name,
        workspace_id = workspace_id,
      )
      stub = nbox_ws_v1.workspace.u(workspace_id).deployments.u(out.id)
      data = stub()
      token = data["deployment"]["api_key"]
      return self.from_serving(
        deployment_id_or_name = out.id,
        token = token,
        workspace_id = workspace_id,
      )

  def map(self, inputs: Union[List[Any], Tuple[Any]]):
    pass

  # /nbx

class Machine():
  def __init__(self, parent_op: Operator, resource: Resource):
    """Simple class to scope the operations to a specific resource"""
    self.parent_op = parent_op
    self.resource = resource

    # these are all the ops that are to be run on this machine as a group
    self.ops_list = []

  def add_child(self, op: Operator):
    self.ops_list.append(op)

  def __enter__(self):
    self.parent_op._current_resource = self.resource

  def __exit__(self, *args):
    self.parent_op._current_resource = None


operator = Operator.fn # convinience



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import google.protobuf.empty_pb2
import grpc
import serve_pb2
import typing

class ServingServiceStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Create: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingRequest,
        serve_pb2.Serving]
    """Creates a new Serving Group.
    If `.models` are specified, the Serving Group will be created with new
    `len(.models)` models.
    """

    GetServing: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingRequest,
        serve_pb2.Serving]
    """Get all available info for a Serving Group. All `.models` are populated."""

    ListServings: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingListRequest,
        serve_pb2.ServingListResponse]
    """Get list of Serving Grpup(s) of a user/workspace."""

    DeleteServing: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingRequest,
        google.protobuf.empty_pb2.Empty]
    """Delete a Serving Group and all its models."""

    UpdateServing: grpc.UnaryUnaryMultiCallable[
        serve_pb2.UpdateServingRequest,
        serve_pb2.Serving]
    """Update a Serving Group. Only the fields specified in the `update_mask` are
    updated.
    """


class ServingServiceServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Create(self,
        request: serve_pb2.ServingRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Serving:
        """Creates a new Serving Group.
        If `.models` are specified, the Serving Group will be created with new
        `len(.models)` models.
        """
        pass

    @abc.abstractmethod
    def GetServing(self,
        request: serve_pb2.ServingRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Serving:
        """Get all available info for a Serving Group. All `.models` are populated."""
        pass

    @abc.abstractmethod
    def ListServings(self,
        request: serve_pb2.ServingListRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.ServingListResponse:
        """Get list of Serving Grpup(s) of a user/workspace."""
        pass

    @abc.abstractmethod
    def DeleteServing(self,
        request: serve_pb2.ServingRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Delete a Serving Group and all its models."""
        pass

    @abc.abstractmethod
    def UpdateServing(self,
        request: serve_pb2.UpdateServingRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Serving:
        """Update a Serving Group. Only the fields specified in the `update_mask` are
        updated.
        """
        pass


def add_ServingServiceServicer_to_server(servicer: ServingServiceServicer, server: grpc.Server) -> None: ...

class ModelServiceStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Watch: grpc.UnaryStreamMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.ServingStatus]
    """Watch for status update of a model in Serving Group."""

    UploadModel: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.Model]
    """Get S3 Upload URL for new/existing Model in Serving Group.
    If id is none then creates new Model with id `id` in the Serving Group.
    If model's serving_group_id is none/doesn't exist, then throws error.
    """

    Deploy: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.Model]
    """Deploy new model in Serving Group."""

    ModelLogs: grpc.UnaryStreamMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.ModelLog]
    """Get logs of a model."""

    ListModels: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelListRequest,
        serve_pb2.ModelListResponse]
    """Get list of models in a Serving Group."""

    DeleteModel: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        google.protobuf.empty_pb2.Empty]
    """Delete a model in a Serving Group."""

    SetModelPin: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        google.protobuf.empty_pb2.Empty]
    """Pinned a model of a Serving Group."""


class ModelServiceServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Watch(self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> typing.Iterator[serve_pb2.ServingStatus]:
        """Watch for status update of a model in Serving Group."""
        pass

    @abc.abstractmethod
    def UploadModel(self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Model:
        """Get S3 Upload URL for new/existing Model in Serving Group.
        If id is none then creates new Model with id `id` in the Serving Group.
        If model's serving_group_id is none/doesn't exist, then throws error.
        """
        pass

    @abc.abstractmethod
    def Deploy(self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Model:
        """Deploy new model in Serving Group."""
        pass

    @abc.abstractmethod
    def ModelLogs(self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> typing.Iterator[serve_pb2.ModelLog]:
        """Get logs of a model."""
        pass

    @abc.abstractmethod
    def ListModels(self,
        request: serve_pb2.ModelListRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.ModelListResponse:
        """Get list of models in a Serving Group."""
        pass

    @abc.abstractmethod
    def DeleteModel(self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Delete a model in a Serving Group."""
        pass

    @abc.abstractmethod
    def SetModelPin(self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Pinned a model of a Serving Group."""
        pass


def add_ModelServiceServicer_to_server(servicer: ModelServiceServicer, server: grpc.Server) -> None: ...



"""
With NimbleBox you can run cluster wide workloads from anywhere. This requires capabilities around distributed computing,
process managements. The code here is tested along with ``nbox.Relic`` to perform distributed local processing.
"""

import os
import inspect
import subprocess
from time import sleep

from threading import Thread
from typing import Union

from nbox import RelicsNBX
from nbox.relics.local import RelicLocal
from nbox import Operator, logger
from nbox.operator import Resource
from nbox.utils import log_traceback

PROC_FOLDER = "./nbx_autogen_proc"


# Manager
class RootRunner(Operator):
  def __init__(
    self,
    op: Operator,
    dist = False,
    relic_type = "nbx",
    verbose: str = False,
    *,
    _unittest: bool = False
  ):
    """This is the Operator which runs on the main Job pod and stores all the values. Think of this like the
    Master node in a distributed system, which contains all the information on the processing."""
    super().__init__()
    self.op = op
    self.dist = dist
    self.verbose = verbose
    self._unittest = _unittest

    _relic_cls = {"local": RelicLocal, "nbx": RelicsNBX}.get(relic_type, None)
    if _relic_cls is None:
      raise ValueError(f"Cannot find relic type: '{relic_type}'")

    self.relic = _relic_cls(relic_name = "hustle_bustle", workspace_id = "wnja9glc", create = True)
    self.cache_status = {}

  def start_status(self):
    def worker():
      # function to keep on updating the cache and seeing the progress
      while True:
        available = 0
        for key in self.keys:
          if self.relic.has(key):
            available += 1
        if self.verbose:
          logger.info(f"Available: [{available}/{len(self.keys)}]")
        sleep(0.69)

    # start the status thread
    self.cache_man = Thread(target = worker, daemon=True)
    self.cache_man.start()

  def prepare_children(self):
    self.op_agents = {}
    self.keys = set()
    for op_name, res in self.op._op_to_resource_map.items():
      if res != None:
        # replace the Opeartors in the graph with AgentsStubs
        op = getattr(self.op, op_name)
        agent_op = AgentOpRootStub(op, op_name, res, self.relic)
        setattr(self.op, op_name, agent_op)
        self.op_agents[op_name] = agent_op
        self.keys.add(op_name + "_input")
        self.keys.add(op_name + "_output")

  def clear_relic(self):
    for op_name in self.op_agents:
      for _x in ["_input", "_output"]:
        key = op_name + _x
        if self.relic.has(key):
          self.relic.rm(key)

  def forward(self, *args, **kwargs):
    if self.dist:
      self.prepare_children()
      if self._unittest:
        self.clear_relic()
      self.start_status()

    return self.op(*args, **kwargs)


class AgentOpRootStub(Operator):
  def __init__(self, op: Operator, op_name:str , res: Resource, relic: Union[RelicLocal, RelicsNBX]):
    """This is the root stub for the remote job. In the current implementation it will spin up a new process
    and manage the connections and things for it.
    
    Hmmm... One day maybe [run go in python](https://medium.com/analytics-vidhya/running-go-code-from-python-a65b3ae34a2d)
    
    Think of this like kubelet or raylet"""
    super().__init__()
    self.op = op
    self.op_name = op_name
    self.relic = relic
    logger.info(f"Initialising AgentOpRootStub: {self.op_name}")

  def get_keys(self):
    input_key = self.op_name +  "_input"
    output_key = self.op_name + "_output"
    return input_key, output_key

  def forward(self, *args, **kwargs):
    # take this op and deploy this to a remote machine where it will start
    # and wait for the remote to finish

    # print(self.relic._objects)
    input_key, output_key = self.get_keys()
    if self.relic.has(output_key):
      self.relic.rm(output_key)
    logger.info(f"{input_key} / {output_key}")

    if not os.path.exists(PROC_FOLDER):
      os.mkdir(PROC_FOLDER)

    fpath = os.path.join(PROC_FOLDER, "auto_" + self.op_name + ".py")
    logger.info(f"Creating file: {fpath}")
    with open(fpath, "w") as f:
      f.write(f'''# Auto generated
import os

os.environ["NBOX_LOG_LEVEL"] = "debug"

from nbox import operator
from nbox.relics import RelicLocal, RelicsNBX
from nbox.utils import load_module_from_path

@operator()
def run():
  # load the operator
  op_cls = load_module_from_path("{self.op.__class__.__qualname__}", "{inspect.getfile(self.op.__class__)}")
  op = op_cls()

  # define the keys
  op_name = "{self.op_name}"
  input_key = "{input_key}"
  output_key = "{output_key}"
  
  # define relic and get things
  relic = RelicsNBX(relic_name = "hustle_bustle", workspace_id = "wnja9glc", create = True)
  args, kwargs = relic.get_object(input_key)
  
  out = op(*args, **kwargs)
  if out == None:
    out = "NBX_NULL"

  # print("Putting", output_key, out)
  relic.put_object(output_key, out)

if __name__ == "__main__":
  # if required user can update the status of the operator
  # run.thaw()
  run()
''')

    # put the data in the object store
    self.relic.put_object(input_key, (list(args), kwargs))

    # run the file as a subprocess: pseudo parallel
    try:
      out = subprocess.run(["python3", fpath], universal_newlines=True)
      if out.returncode != 0:
        raise Exception(f"Process returned non zero code: {out.returncode}")
    except KeyboardInterrupt:
      logger.info("KeyboardInterrupt, closing connections")
    except Exception as e:
      logger.error(f"Error in process: '{self.op_name}', check above for details")
      log_traceback()
      raise e

    # get the results from the object store
    out = self.relic.get_object(output_key)
    while out == None:
      sleep(1)
      out = self.relic.get_object(output_key)
    return out




"""
All the operators that are using the demos are provided here. Highly recommend someome to go through
this when learning how to code your own ``Operator``.
"""

from nbox import Operator, logger
from nbox.lib.shell import ShellCommand, GitClone

class Intro(Operator):
  # yet another time waster
  def __init__(self):
    super().__init__()

  def forward(self):
    logger.info("Welcome to NimbleBox!")


class ThingsShellCanDo(Operator):
  def __init__(self) -> None:
    super().__init__()
    self.amar = ShellCommand("echo 'Lana Del Ray'")
    self.akbar = ShellCommand("echo 'Run any shell command, like read files'")
    self.anthony = ShellCommand("echo 'Kenrik'")
  
  def forward(self):
    self.amar()
    self.akbar()
    self.anthony()


class Magic(Operator):
  # named after the Dr.Yellow train that maintains the Shinkansen
  def __init__(self, strength = "very"):
    super().__init__()
    self.strength = strength
    self.intro = Intro()
    self.shell_demo = ThingsShellCanDo()
    # self.cloner = GitClone("https://github.com/yashbonde/yQL")

  def forward(self):
    # self.cond1(cond = 1)
    self.intro()
    self.shell_demo()
    # self.cloner()


class MagicServing(Operator):
  def __init__(self):
    super().__init__()
    self.intro = Intro()
    self.cities = [
      "Mumbai",
      "Delhi",
      "Banares",
      "Bangalore",
      "Chennai",
      "Kolkata",
      "Hyderabad",
      "Ahmedabad",
      "Pune",
      "Jaipur",
      "Surat",
      "Lucknow",
      "Kanpur",
      "Nagpur",
      "Patna",
      "Indore",
      "Vadodara",
      "Agra",
      "Siliguri",
      "Coimbatore",
    ]

  def __remote_init__(self):
    logger.info("Running remote init: it will only run at initialisation")
    logger.info("So you can do things like:")
    logger.info("    * initialise your model")
    logger.info("    * connect to databases")
    logger.info("    * create games!")
    self.intro()

  def forward(self, x: str = ""):
    import random
    if not x:
      return {"data": f"Hi, I am {random.choice(self.cities)}!"}
    else:
      return {"data": f"Echo back: {x}!"}
